from textual import events
from textual.app import ComposeResult
from textual.widgets import Button
from textual.widgets import Static

# Tooltips

BTN_SET_FPGA_DEFAULTS = """
Press this button to set the FPGA default values for the F-FEE DEB and AEB units. 

This command will take about 5 cycles: one cycle for settings the DEB defaults, and an additional cycle for each AEB unit.

"""

class GeneralCommand(Static):

    def compose(self) -> ComposeResult:
        yield Button("Set FPGA Defaults", id='btn-set-fpga-defaults', classes='command')

    def _on_mount(self, event: events.Mount) -> None:
        self.query_one("#btn-set-fpga-defaults").tooltip = BTN_SET_FPGA_DEFAULTS



import pickle
import time
from typing import Tuple

import zmq
from egse.zmq import MessageIdentifier


def log_message(app, name, sync_id, message, timed_out: bool):
    # app.log(f"{name:10s} {'' if sync_id is None else MessageIdentifier(sync_id).name}, {'🔴' if timed_out else '🟢'}")
    ...


services = {
    'cm_cs': {
        'description': "The Configuration Control Server",
        'hostname': 'localhost',
        'port': 6001,
        'type': zmq.SUB,
        'multipart': False,
        'sock': None,
        'interval': 3,
        'last_received': time.monotonic(),
        'timed-out': False,
        'callback': log_message
    },
    'sm_cs': {
        'description': "The Storage Manager",
        'hostname': 'localhost',
        'port': 6101,
        'type': zmq.SUB,
        'multipart': False,
        'sock': None,
        'interval': 3,
        'last_received': time.monotonic(),
        'timed-out': False,
        'callback': log_message
    },
    'pm_cs': {
        'description': "The Process Manager",
        'hostname': 'localhost',
        'port': 6201,
        'type': zmq.SUB,
        'multipart': False,
        'sock': None,
        'interval': 3,
        'last_received': time.monotonic(),
        'timed-out': False,
        'callback': log_message
    },
    'syn_cs': {
        'description': "The Synoptics Control Server storing generic and device independent housekeeping.",
        'hostname': 'localhost',
        'port': 6205,
        'type': zmq.SUB,
        'multipart': False,
        'sock': None,
        'interval': 3,
        'last_received': time.monotonic(),
        'timed-out': False,
        'callback': log_message
    },
    # The data dumper sends out a STATUS sync_id every second (because the timeout of the Poller is set to 1000ms).
    "data_dump": {
        'description': "The Data Dumper for F-CAM SpaceWire data.",
        'hostname': 'localhost',
        'port': 30304,  # MONITORING_PORT
        'type': zmq.SUB,
        'subscribe': MessageIdentifier.STATUS.to_bytes(1, byteorder='big'),
        'multipart': True,
        'sock': None,
        'interval': 3,
        'last_received': time.monotonic(),
        'timed-out': False,
        'callback': log_message
    },
    "dpu_cs": {
        'description': "The DPU Processor",
        'hostname': 'localhost',
        'port': 6601,  # MONITORING_PORT
        'type': zmq.SUB,
        'multipart': False,
        'sock': None,
        'interval': 3,
        'last_received': time.monotonic(),
        'timed-out': False,
        'callback': log_message
    },
    # Monitor DATA_DISTRIBUTION port, this data is sent out by the Data Processor,
    # Subscribe to the SYNC_TIMECODE which is sent out every 2.5s for the F-CAM. Other sync identifiers are ignored.
    "data": {
        'description': "The Data Distribution by the Data Processor",
        'hostname': 'localhost',
        'port': 30103,  # DATA_DISTRIBUTION_PORT
        'type': zmq.SUB,
        'subscribe': MessageIdentifier.SYNC_TIMECODE.to_bytes(1, byteorder='big'),
        'multipart': True,
        'sock': None,
        'interval': 5,
        'last_received': time.monotonic(),
        'timed-out': False,
        'callback': log_message
    },
}


async def handle_multi_part(sock: zmq.Socket, message_id: bytes) -> Tuple[int, list]:
    message_parts = []
    message_id = int.from_bytes(message_id, byteorder='big')
    while True:
        part = await sock.recv()
        message_parts.append(pickle.loads(part))
        if not sock.getsockopt(zmq.RCVMORE):
            break

    return message_id, message_parts


async def handle_single_part(sock: zmq.Socket, message: bytes) -> Tuple[int, list]:
    message_id = MessageIdentifier.ALL
    response = pickle.loads(message)

    return message_id, [response]



from __future__ import annotations

import logging
import pickle
import queue
import random
import threading
import time
import traceback
from queue import Queue
from typing import TYPE_CHECKING

import zmq
from egse.dpu.fdpu import FastCameraDPUProxy
from egse.fee.ffee import HousekeepingData
from egse.fee.ffee import aeb_state
from egse.fee.ffee import f_fee_mode
from egse.reg import RegisterMap
from egse.settings import Settings
from egse.setup import load_setup
from egse.zmq import MessageIdentifier

from .messages import AebStateChanged
from .messages import DebModeChanged
from .messages import DtcInModChanged
from .messages import ExceptionCaught
from .messages import OutbuffChanged
from .messages import TimeoutReached

if TYPE_CHECKING:
    from .app import FastFEEApp

_LOGGER = logging.getLogger("egse.f-fee-tui")

dpu = Settings.load("DPU Processor")


class Command(threading.Thread):
    def __init__(self, app: 'FastFEEApp', command_q: Queue) -> None:
        super().__init__()
        self._app = app
        self._command_q = command_q
        self._f_dpu: FastCameraDPUProxy | None = None
        self._canceled = threading.Event()

    def run(self) -> None:

        with FastCameraDPUProxy() as self._f_dpu:
            while True:
                if self._canceled.is_set():
                    break
                try:
                    target, command, args, kwargs = self._command_q.get_nowait()
                    self.execute_command(target, command, args, kwargs)
                    self._command_q.task_done()
                except queue.Empty:
                    time.sleep(0.1)  # Sleep briefly to avoid busy-waiting
                    continue
                except Exception as exc:
                    _LOGGER.error(f"Caught and exception: {exc}")
                    tb = traceback.format_exc()
                    self._app.post_message(ExceptionCaught(exc, tb))

    def cancel(self) -> None:
        self._canceled.set()

    def execute_command(self, target: str, command: str, args: list, kwargs: dict):
        if target == "DPU":
            try:
                return getattr(self._f_dpu, command)(*args, **kwargs)
            except AttributeError as exc:
                _LOGGER.error(f"No such command for DPU: {command}", exc_info=True)


class Monitor(threading.Thread):
    def __init__(self, app: 'FastFEEApp') -> None:
        self._app = app
        self._canceled = threading.Event()
        self._reset_frame_errors = threading.Event()

        self.hostname = dpu.HOSTNAME
        self.port = dpu.DATA_DISTRIBUTION_PORT

        self.previous_deb_mode = f_fee_mode.ON_MODE
        self.previous_aeb_state = {}
        self.accumulated_outbuff = [0, 0, 0, 0, 0, 0, 0, 0]
        self.outbuff_mapping = [0, 2, 1, 3, 4, 6, 5, 7]
        """Mapping of the OUTBUFF_x and the DTC_IN_MOD, se Setup AEB_TO_T_IN_MOD"""

        super().__init__()

    def run(self) -> None:

        n_timeouts = 0
        """The timeout is 1s, we count the number of timeouts to detect if the DPU or N-FEE might be dead or stalled."""

        context = zmq.Context.instance()
        receiver = context.socket(zmq.SUB)
        receiver.setsockopt_string(zmq.SUBSCRIBE, MessageIdentifier.F_FEE_REGISTER_MAP.to_bytes(1, byteorder='big'))
        receiver.setsockopt_string(zmq.SUBSCRIBE, MessageIdentifier.SYNC_HK_DATA.to_bytes(1, byteorder='big'))
        receiver.connect(f"tcp://{self.hostname}:{self.port}")

        setup = load_setup()

        while True:
            if self._canceled.is_set():
                break
            if self._reset_frame_errors.is_set():
                self._reset_frame_errors.clear()
                self.accumulated_outbuff = [0, 0, 0, 0, 0, 0, 0, 0]
                self._app.post_message(OutbuffChanged(self.accumulated_outbuff))

            socket_list, _, _ = zmq.select([receiver], [], [], timeout=1.0)

            if receiver in socket_list:
                n_timeouts = 0
                try:
                    sync_id, pickle_string = receiver.recv_multipart()
                    sync_id = int.from_bytes(sync_id, byteorder='big')
                    data = pickle.loads(pickle_string)
                    self.handle_messages(sync_id, data, setup)
                except Exception as exc:
                    tb = traceback.format_exc()
                    self._app.post_message(ExceptionCaught(exc, tb))

            if len(socket_list) == 0:
                n_timeouts += 1
                if n_timeouts > 3:  # at least a timecode should arrive every 2.5s
                    self._app.post_message(TimeoutReached("Timeout reached after 3s on monitoring channel."))
                    n_timeouts = 0

        receiver.disconnect(f"tcp://{self.hostname}:{self.port}")
        receiver.close()

    def cancel(self) -> None:
        self._canceled.set()

    def reset_frame_errors(self):
        self._reset_frame_errors.set()

    def handle_messages(self, sync_id, data, setup):

        if sync_id == MessageIdentifier.F_FEE_REGISTER_MAP:

            # How can we be sure the Register Map is properly synchronised with the F-FEE?

            register_map, _ = data
            register_map = RegisterMap("F-FEE", memory_map=register_map, setup=setup)

            t0 = register_map["DEB_DTC_IN_MOD_2", "T0_IN_MOD"]
            t1 = register_map["DEB_DTC_IN_MOD_2", "T1_IN_MOD"]
            t2 = register_map["DEB_DTC_IN_MOD_2", "T2_IN_MOD"]
            t3 = register_map["DEB_DTC_IN_MOD_2", "T3_IN_MOD"]
            t4 = register_map["DEB_DTC_IN_MOD_1", "T4_IN_MOD"]
            t5 = register_map["DEB_DTC_IN_MOD_1", "T5_IN_MOD"]
            t6 = register_map["DEB_DTC_IN_MOD_1", "T6_IN_MOD"]
            t7 = register_map["DEB_DTC_IN_MOD_1", "T7_IN_MOD"]

            self._app.log(f"{t0=}, {t1=}, {t2=}, {t3=}, {t4=}, {t5=}, {t6=}, {t7=}")

            self._app.post_message(DtcInModChanged(t0, t1, t2, t3, t4, t5, t6, t7))

        elif sync_id == MessageIdentifier.SYNC_HK_DATA:

            cmd, aeb_id, data, timestamp = data

            if cmd == 'command_deb_read_hk':
                hk_data = HousekeepingData("DEB", data, setup)
                deb_mode = f_fee_mode(hk_data["STATUS", "OPER_MOD"])
                self._app.log(f"DEB_MODE = {deb_mode.name}")
                self._app.post_message(DebModeChanged(deb_mode))
                outbuff = [hk_data["OVF", f"OUTBUFF_{x}"] for x in range(1, 9)]
                # outbuff = random.choices([0, 1], k=8)
                self._app.log(f"{outbuff = }")
                if any(outbuff):
                    # re-order outbuff to match with DTC_IN_MOD
                    outbuff = [outbuff[idx] for idx in self.outbuff_mapping]
                    # update the accumulated values for OUTBUFF
                    self.accumulated_outbuff = [x + y for x, y in zip(self.accumulated_outbuff, outbuff)]
                    self._app.post_message(OutbuffChanged(self.accumulated_outbuff))
            elif cmd == 'command_aeb_read_hk':
                aeb_id = aeb_id[0]  # this comes from the args, so it's a list
                hk_data = HousekeepingData(aeb_id, data, setup)
                aeb_status = hk_data["STATUS", "AEB_STATUS"]
                self._app.log(f"AEB_STATE = {aeb_state(aeb_status).name}")

                # Power in handled by the DEB_DTC_AEB_ONOFF state from the DEB register map above.

                if aeb_status == aeb_state.OFF:
                    self._app.post_message(AebStateChanged(f"{aeb_id.lower()}_onoff", False))
                elif aeb_status == aeb_state.INIT:
                    self._app.post_message(AebStateChanged(f"{aeb_id.lower()}_init", True))
                elif aeb_status == aeb_state.POWER_UP:
                    self._app.post_message(AebStateChanged(f"{aeb_id.lower()}_power_up", True))
                elif aeb_status == aeb_state.POWER_DOWN:
                    self._app.post_message(AebStateChanged(f"{aeb_id.lower()}_power_down", True))
                elif aeb_status == aeb_state.CONFIG:
                    self._app.post_message(AebStateChanged(f"{aeb_id.lower()}_config", True))
                elif aeb_status == aeb_state.IMAGE:
                    self._app.post_message(AebStateChanged(f"{aeb_id.lower()}_image", True))
                elif aeb_status == aeb_state.PATTERN:
                    self._app.post_message(AebStateChanged(f"{aeb_id.lower()}_pattern", True))



import itertools

from textual.app import ComposeResult
from textual.widgets import Label
from textual.widgets import Sparkline
from textual.widgets import Static

from f_fee_tui.messages import DtcInModChanged

ON = "✖"
OFF = ""

T0 = {0b001: "T0-001"}
T1 = {0b001: "T1-001", 0b010: "T1-010"}
T2 = {0b001: "T2-001", 0b010: "T2-010"}
T3 = {0b001: "T3-001"}
T4 = {0b001: "T4-001"}
T5 = {0b001: "T5-001", 0b010: "T5-010"}
T6 = {0b001: "T6-001", 0b010: "T6-010"}
T7 = {0b001: "T7-001"}

ALL_IDS = list(itertools.chain(
        T0.values(), T1.values(), T2.values(), T3.values(), T4.values(), T5.values(), T6.values(), T7.values()
))


class DtcInMod(Static):
    """A widget to monitor the state of the AEBs."""

    def compose(self) -> ComposeResult:
        yield Label("", classes="header")
        yield Label("AEB1", classes="two-cols header")
        yield Label("AEB2", classes="two-cols header")
        yield Label("AEB3", classes="two-cols header")
        yield Label("AEB4", classes="two-cols header")

        yield Label("SpW 1")
        yield Label("", id="T0-001", classes="one-col")    # AEB1-F
        yield Label("", id="T1-001", classes="one-col")    # AEB1-E
        yield Label("", id="T1-010", classes="one-col")    # AEB2-F
        yield Label("", classes="one-col disabled")
        yield Label("", classes="one-col disabled")
        yield Label("", classes="one-col disabled")
        yield Label("", classes="one-col disabled")
        yield Label("", classes="one-col disabled")

        yield Label("SpW 2")
        yield Label("", classes="one-col disabled")
        yield Label("", id="T2-010", classes="one-col")    # AEB1-E
        yield Label("", id="T2-001", classes="one-col")    # AEB2-F
        yield Label("", id="T3-001", classes="one-col")    # AEB2-E
        yield Label("", classes="one-col disabled")
        yield Label("", classes="one-col disabled")
        yield Label("", classes="one-col disabled")
        yield Label("", classes="one-col disabled")

        yield Label("SpW 3")
        yield Label("", classes="one-col disabled")
        yield Label("", classes="one-col disabled")
        yield Label("", classes="one-col disabled")
        yield Label("", classes="one-col disabled")
        yield Label("", id="T4-001", classes="one-col")    # AEB3-F
        yield Label("", id="T5-001", classes="one-col")    # AEB3-E
        yield Label("", id="T5-010", classes="one-col")    # AEB4-F
        yield Label("", classes="one-col disabled")

        yield Label("SpW 4")
        yield Label("", classes="one-col disabled")
        yield Label("", classes="one-col disabled")
        yield Label("", classes="one-col disabled")
        yield Label("", classes="one-col disabled")
        yield Label("", classes="one-col disabled")
        yield Label("", id="T6-010", classes="one-col")    # AEB3-E
        yield Label("", id="T6-001", classes="one-col")    # AEB4-F
        yield Label("", id="T7-001", classes="one-col")    # AEB4-E

        yield Label("", classes="footer")
        yield Label("F", classes="one-col footer")
        yield Label("E", classes="one-col footer")
        yield Label("F", classes="one-col footer")
        yield Label("E", classes="one-col footer")
        yield Label("F", classes="one-col footer")
        yield Label("E", classes="one-col footer")
        yield Label("F", classes="one-col footer")
        yield Label("E", classes="one-col footer")

        yield Label("Errors", id="lbl-frame-errors")
        yield Sparkline([0, 0, 0, 0, 0, 0, 0, 0], id="frame-errors")

    def on_mount(self) -> None:
        tooltip_msg = "Accumulated errors on SpW transmit buffer overflow."
        self.query_one("#lbl-frame-errors", Label).tooltip = tooltip_msg
        self.query_one("#frame-errors", Sparkline).tooltip = tooltip_msg

    def set_state(self, state: DtcInModChanged):
        self.log(f"{state.t0=}, {state.t1=}, {state.t2=}, {state.t3=}, {state.t4=}, {state.t5=}, {state.t6=}, {state.t7=}")
        for id_ in ALL_IDS:
            self.query_one(f"#{id_}", Label).update(OFF)

        if state.t0 == 0b001:
            self.query_one("#T0-001", Label).update(ON)

        if state.t1 == 0b001:
            self.query_one("#T1-001", Label).update(ON)
        elif state.t1 == 0b010:
            self.query_one("#T1-010", Label).update(ON)

        if state.t2 == 0b001:
            self.query_one("#T2-001", Label).update(ON)
        elif state.t2 == 0b010:
            self.query_one("#T2-010", Label).update(ON)

        if state.t3 == 0b001:
            self.query_one("#T3-001", Label).update(ON)

        if state.t4 == 0b001:
            self.query_one("#T4-001", Label).update(ON)

        if state.t5 == 0b001:
            self.query_one("#T5-001", Label).update(ON)
        elif state.t5 == 0b010:
            self.query_one("#T5-010", Label).update(ON)

        if state.t6 == 0b001:
            self.query_one("#T6-001", Label).update(ON)
        elif state.t6 == 0b010:
            self.query_one("#T6-010", Label).update(ON)

        if state.t7 == 0b001:
            self.query_one("#T7-001", Label).update(ON)



from textual.app import ComposeResult
from textual.widgets import Static

from .leds import OnOffLedWithLabel


class DEBMode(Static):
    """A widget to monitor the DEB mode."""

    def compose(self) -> ComposeResult:
        yield OnOffLedWithLabel("ON", True, id="deb-on")
        yield OnOffLedWithLabel("STANDBY", False, id="deb-standby")
        yield OnOffLedWithLabel("FULL_IMAGE", False, id="deb-full-image")
        yield OnOffLedWithLabel("FULL_IMAGE_PATTERN", False, id="deb-full-image-pattern")
        yield OnOffLedWithLabel("WINDOWING", False, id="deb-windowing")
        yield OnOffLedWithLabel("WINDOWING_PATTERN", False, id="deb-windowing-pattern")



from typing import Tuple

from textual.app import ComposeResult
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import Label
from textual.widgets import Static


class OnOffLed(Static):
    """A Label widget that represents a LED with two states: ON and OFF."""

    state = reactive(False)

    def __init__(self, **kwargs):
        self._true, self._false = kwargs.pop("states", ("🟩", "🟥"))
        super().__init__(**kwargs)

    def watch_state(self, state: bool) -> None:
        # self.update("🟢" if self.state else "🔴")
        # self.update("✅" if self.state else "")
        # self.update("🟩" if state else "🟥")
        self.update(self._true if self.state else self._false)


class OnOffLedWithLabel(Widget):
    """A widget with a label and a LED with ON/OFF state."""

    def __init__(self, label: str, state: bool, id: str = None) -> None:
        self.label = label
        self._state = state
        super().__init__(id=id)

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, state: bool):
        self._state = state
        self.led.state = state

    def compose(self) -> ComposeResult:
        self.led = OnOffLed()
        self.led.state = self.state

        yield Label(self.label)
        yield self.led



from textual.message import Message


class DebModeChanged(Message):
    """This message is sent when the Monitor detects a change in the DEB mode."""
    def __init__(self, deb_mode: int):
        super().__init__()
        self.deb_mode = deb_mode


class AebStateChanged(Message):
    """This message reports on the State of the AEBs."""
    def __init__(self, aeb_state_type, aeb_state):
        super().__init__()
        self.aeb_state_type = aeb_state_type
        self.aeb_state = aeb_state


class DtcInModChanged(Message):
    """This message reports on the State of the AEBs."""
    def __init__(self, t0, t1, t2, t3, t4, t5, t6, t7):
        super().__init__()
        self.t0 = t0
        self.t1 = t1
        self.t2 = t2
        self.t3 = t3
        self.t4 = t4
        self.t5 = t5
        self.t6 = t6
        self.t7 = t7


class OutbuffChanged(Message):
    def __init__(self, outbuff):
        super().__init__()
        self.outbuff = outbuff


class ExceptionCaught(Message):
    """This message is sent whenever a non-resolvable exception occurs in the Monitor or Commanding thread."""
    def __init__(self, exc: Exception, tb=None):
        super().__init__()
        self.exc = exc
        self.tb = tb


class TimeoutReached(Message):
    """This message is sent when the Monitor or Commanding thread reached a timeout."""
    def __init__(self, msg: str):
        super().__init__()
        self.message = msg


class ProblemDetected(Message):
    """A notification message for a problem that needs reporting."""
    def __init__(self, msg: str,):
        super().__init__()
        self.message = msg
